/* sqtile.c -- Square-Figure tiling functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqtile.c,v 1.2 2003/03/18 07:59:45 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqtile.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* local stuff */
static int plotter(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension);
static int printBackStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension);
static int printForwardStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension);
static int stop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension);
static int tilingEngine(
	figureSet_t const* figureSet,
	int figures, /* figureSet->siblings->size */
	int figureSize,
	int xDimension,
	int yDimension, /* boardSize / xDimension */
	int* tiling,
	int* board,
	int boardSize, /* figures * figureSize */
	int* taken,
	int continuous,
	int next, /* first or next */
	int (*found)( /* found callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension),
	int (*forward)( /* forward info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension),
	int (*back)( /* back info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension));
static int tilingEngineHelper(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int next);


/** Figure/sibling/position tuple.  s is (and must be!) smaller than 8
* [maximal number of siblings in a sibling group, symmetry property].
* p is (and must be!) smaller than 16 [implementation limitation imposed
* by sqfig].  (f<<3|s)<<4|p is a more performant version for
* (f*8+s)*16+p using the fact 8 and 16 are powers of 2.
*/
#define FST(f, s, p) (((f) << 3 | (s)) << 4 | (p))

/** Figure from figure/sibling/position tuple.  More performant version
* for (t/16)/8.
*/
#define FIG(t) ((t) >> 7)

/** Sibling from figure/sibling/position tuple.  More performant version
* for (t/16)%8.
*/
#define SIB(t) ((t) >> 4 & 0x7)

/** Position from figure/sibling/position tuple.  More performant
* version for t%16.
*/
#define POS(t) ((t) & 0xf)


/** Get first tiling for a given X dimension.  The tiling buffer must be
* able to hold DISTINCT_FIGURE_COUNT(figureSet) ints, which will carry
* on return the figure indices.  The board buffer must be able to hold
* figureSize*DISTINCT_FIGURE_COUNT(figureSet) ints and will carry on
* return the board's tiling layout as indices into the tiling buffer.
* The return value indicates if a solution has been found (1) or not (0).
*/
int getFirstTiling(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous)
{
	return tilingEngineHelper(
		figureSet,
		xDimension,
		tiling,
		board,
		continuous,
		0);
}


/** Get next tiling for a given X dimension.  The tiling buffer must be
* able to hold DISTINCT_FIGURE_COUNT(figureSet) ints, which must have
* been filled by a previous succeeding getFirstTiling() call and will
* carry on return the figure indices.  The board buffer must be able to
* hold figureSize*DISTINCT_FIGURE_COUNT(figureSet) ints and will carry
* on return the board's tiling layout as indices into the tiling buffer.
* The return value indicates if a solution has been found (1) or not (0).
*/
int getNextTiling(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous)
{
	return tilingEngineHelper(
		figureSet,
		xDimension,
		tiling,
		board,
		continuous,
		1);
}


/** Tiling engine helper.
*/
static int tilingEngineHelper(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int next)
{
	int figures, figureSize, boardSize, yDimension, r;

	/* transient data structure that needs to be deallocated upon
	* return from this function
	*/
	int* taken;

	/* number of (distinct) figures */
	figures = figureSet->siblings->size;
	if (figures < 1) {
		/* Empty figure list means zero sized figure
		* and hence zero sized tiling buffer and board buffer.
		* We won't handle this degenerate case (especially since
		* due to zero buffer size there is no place to keep state
		* for getNextTiling()), so we just return 0.
		*/
		return 0; /* not found */
	}

	figureSize = figureSet->siblings->siblings[0]->figures[0].size;
	boardSize = figures * figureSize;

	/* check the appropriateness of xDimension */
	if (xDimension == 0 || boardSize % xDimension != 0) {
		/* Can't generate a rectangle with xDimension,
		* that means no solution.
		*/
		return 0; /* not found */
	}
	yDimension = boardSize / xDimension;

	/* allocate and initialize transient stuff */
	taken = malloc(figures * sizeof(*taken));
	if (taken == NULL) {
		return 0; /* can't distinguish 'failed' from 'not found' */
	}

	r = tilingEngine(
		figureSet,
		figures,
		figureSize,
		xDimension,
		yDimension,
		tiling,
		board,
		boardSize,
		taken,
		continuous,
		next, /* first or next */
		stop, /* stop if first or next is found */
		NULL, /* or printForwardStatus, plotter */
		NULL); /* or printBackStatus */

	free(taken);

	return r;
}


/** Find the next tiling.
*/
static int tilingEngine(
	figureSet_t const* figureSet,
	int figures, /* figureSet->siblings->size */
	int figureSize,
	int xDimension,
	int yDimension, /* boardSize / xDimension */
	int* tiling,
	int* board,
	int boardSize, /* figures * figureSize */
	int* taken,
	int continuous,
	int next, /* first or next */
	int (*found)( /* found callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension),
	int (*forward)( /* forward info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension),
	int (*back)( /* back info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension))
{
	int current, boardNext;
	int i, j, r, fig, sib, pos, sibSize, x, y, xx, yy;
	figure_t f;
	square_t s;
	int nb, nbc[MAXFIGURESIZE];

	/* check the special case -- we need at least one array element */
	if (figures <= 0) {
		return 0;
	}

	if (next) {
		/* setup to find next solution */

		/* 'board' must be set up from earlier invocations */

		boardNext = boardSize;

		for (i = 0; i < figures; i++) {
			taken[i] = 1;
		}

		current = figures;

		/* 'tiling' must be set up from earlier invocations */

		goto backtrack;
	}

	/* setup to find first solution */

	/* Initialize board.  Any of board[i] must not be a valid index
	* into 'tiling' (well, except for the last one, figures-1).
	*/
	for (i = 0; i < boardSize; i++) {
		board[i] = -1;
	}

	/* next free entry in board, to suppress need for linear search */
	boardNext = 0;

	/* init 'taken' to untaken */
	for (i = 0; i < figures; i++) {
		taken[i] = 0;
	}

	current = 0;

	tiling[current] = FST(0, 0, 0); /* we asserted earlier that figures > 0 */

	/* Note: 'current' is advanced forth and back until a solution is
	* found or none (more) can be found.  [backtracking]
	*/
	for (;;) {

		/* invariants:
		* o 'tiling' is ok up to and including 'current - 1' (this is true
		*   too in the initial special case, 'current == 0')
		* o 'current < figures', i.e. we're not yet done
		* o 'tiling[current]' points on the next possible candidate, that
		*   is however not necessary a valid candidate (it may either be an
		*   already taken figure or not fit)
		*/

		/* search matching figure @current */

		/* find next figure candidate that is not taken;
		* [faster lookup instead of linear search in 'tiling']
		*/
		while (taken[FIG(tiling[current])]) {
nextfig :
			/* next figure */

			fig = FIG(tiling[current]) + 1;
			tiling[current] = FST(fig, 0, 0);

			/* check if we overshot it */
			if (fig == figures) {
				goto backtrack;
			}
		}

		/* find a matching sibling */

		for (;;) {

			/* init these variables here since we're
			* jumping right into the code when backtracking, etc
			*/

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;
			pos = POS(tiling[current]);

			/* 'boardNext' is the smallest free entry;
			* 'f.squares' is ordered the same way 'board' is,
			* which can be used in the non-continuous mode
			*/
			x = boardNext % xDimension;
			y = boardNext / xDimension;

			/* note: the non-continuous mode needs only to consider
			* f.squares[0] since the other squares will colide with
			* stuff already being on the board, do to their ordering
			*/

			FIGURE_COPY(&f, &figureSet->siblings->siblings[fig]->figures[sib]);
			SQUARE_COPY(&s, &f.squares[pos]);

			/* normalize to the square, if needed */
			if (s.x != 0 || s.y != 0) {
				for (i = 0; i < f.size; i++) {
					f.squares[i].x = (squareCoordinate_t)(f.squares[i].x - s.x);
					f.squares[i].y = (squareCoordinate_t)(f.squares[i].y - s.y);
				}
			}

			/* check for match */

			/* separate code for speedup */
			if (!continuous) {

				for (i = 0; i < f.size; i++) {

					xx = x + f.squares[i].x;
					if (xx < 0 || xx >= xDimension) {
						goto nextsib;
					}

					yy = y + f.squares[i].y;
					if (yy < 0 || yy >= yDimension) {
						goto nextsib;
					}

					/* check for availability */

					nb = xx + yy * xDimension;
					if (board[nb] != -1) {
						goto nextsib;
					}

					/* cache [for speedup] */

					nbc[i] = nb;
				}

			} else {

				for (i = 0; i < f.size; i++) {

					/* we don't handle continuation with rotated
					* or mirrored figures
					*/

					/* we don't handle continuation over more than
					* one cell
					*/

					xx = x + f.squares[i].x;
					if (xx < 0) {
						xx += xDimension;
						if (xx < 0) {
							goto nextpos;
						}
					} else if (xx >= xDimension) {
						xx -= xDimension;
						if (xx >= xDimension) {
							goto nextpos;
						}
					}

					yy = y + f.squares[i].y;
					if (yy < 0) {
						yy += yDimension;
						if (yy < 0) {
							goto nextpos;
						}
					} else if (yy >= yDimension) {
						yy -= yDimension;
						if (yy >= yDimension) {
							goto nextpos;
						}
					}

					nb = xx + yy * xDimension;
					if (board[nb] != -1) {
						goto nextpos;
					}

					nbc[i] = nb;
				}

				/* check if two (or more) 'wrapped' squares
				* intend to occupy the same board position;
				* reject if so
				*/
				/* linear search on unordered data, however
				* the array has only up to 16 elements
				*/
				for (i = 0; i < f.size - 1; i++) {
					for (j = i + 1; j < f.size; j++) {
						if (nbc[i] == nbc[j]) {
							goto nextpos;
						}
					}
				}
			}

			taken[fig] = 1;

			/* update board */
			for (i = 0; i < f.size; i++) {
				board[nbc[i]] = current;
			}

			/* update 'boardNext'; linear search */
			while (boardNext < boardSize && board[boardNext] != -1) {
				boardNext++;
			}

			break;

backtrack :
			/* backtrack */

			current--;

			/* back info callback */
			if (back != NULL) {
				r = (*back)(
					tiling,
					figures,
					current,
					board,
					xDimension,
					yDimension);
				if (r) {
					return 0;
				}
			}

			if (current < 0) {
				/* done, not found */
				/* an info callback [with current == 0] has already been done */
				return 0;
			}

			/* reset 'taken' */
			taken[FIG(tiling[current])] = 0;

			/* Reset board; we can't leave 'current' since
			* in a later stage that would be indistinguishable
			* from newer such values.  This is suboptimally
			* a linear search.
			*/
			for (i = 0; i < boardSize; i++) {
				if (board[i] == current) {
					board[i] = -1;
					/* possibly update 'boardNext' */
					if (i < boardNext) {
						boardNext = i;
					}
				}
			}

			/* don't reset tiling[current]; search on from there */

nextpos :
			/* next position */

			if (!continuous) {
				goto nextsib;
			}

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;
			pos = POS(tiling[current]);

			pos++;
			if (pos < figureSize) {
				tiling[current] = FST(fig, sib, pos);
				continue;
			}

nextsib :
			/* next sibling */

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;

			sib++;
			if (sib == sibSize) {
				goto nextfig;
			}

			tiling[current] = FST(fig, sib, 0);
		}

		/* next step */
		current++;

		/* to trace the search,
		* plotTiling(board, xDimension, yDimension, 0, 1)
		* could be called here, established e.g. by the
		* 'forward' callback
		*/

		/* forward info callback */
		if (forward != NULL) {
			r = (*forward)(
				tiling,
				figures,
				current,
				board,
				xDimension,
				yDimension);
			if (r) {
				return 0;
			}
		}

		if (current == figures) {
			/* done, found */

			/* found callback */
			if (found != NULL) {
				r = (*found)(
					tiling,
					figures,
					current,
					board,
					xDimension,
					yDimension);
				/* callback's return value controls flow */
				if (r) {
					return 1; /* ok */
				}

			} else {
				return 1; /* ok */
			}
		}

		/* reset that next 'current', start that search @zero */
		tiling[current] = FST(0, 0, 0);
	}
}


/** Plot a tiling.
*/
void plotTiling(
	int const* board,
	int xDimension,
	int yDimension,
	int plotDots,
	int hatchUnassigned)
{
	int x, y, i, a, b, c, d;

	/* must be at least 1*1, smaller will be 'empty' */
	if (xDimension <= 0 || yDimension <= 0) {
		return;
	}

	/* top row edge */

	y = yDimension - 1;
	printf("+"); /* top left */
	printf("--");

	for (x = 1; x < xDimension; x++) {
		/* depends on previous x and current x */
		a = board[(x - 1) + y * xDimension];
		b = board[x + y * xDimension];

		if (a != b) {
			printf("+");
		} else {
			printf("-");
		}
		printf("--");
	}

	printf("+"); /* top right */
	printf("\n");

	/* main body */
	for (; y > 0; y--) {

		/* contents */
		for (i = 0; i < 2; i++) { /* [size 2] */
			x = 0;
			b = board[x + y * xDimension];

			printf("|"); /* left */

			if (hatchUnassigned && b == -1) {
				printf("\\\\");
			} else {
				printf("  ");
			}

			for (x++; x < xDimension; x++) {
				/* depends on previous x and current x */
				a = board[(x - 1) + y * xDimension];
				b = board[x + y * xDimension];

				if (a != b) {
					printf("|");
				} else if (plotDots) {
					printf(".");
				} else if (hatchUnassigned && b == -1) {
					printf("\\");
				} else {
					printf(" ");
				}

				if (hatchUnassigned && b == -1) {
					printf("\\\\");
				} else {
					printf("  ");
				}
			}

			printf("|");
			printf("\n");
		}

		/* possible delimiters */
		/* left, depends on current y and next y */
		x = 0;
		b = board[x + y * xDimension];
		d = board[x + (y - 1) * xDimension];

		if (b != d) {
			printf("+");
			printf("--");
		} else {
			printf("|");
			if (plotDots) {
				printf("..");
			} else if (hatchUnassigned && b == -1) {
				printf("\\\\");
			} else {
				printf("  ");
			}
		}

		for (x = 1; x < xDimension; x++) {
			/* depends on four fields */
			a = board[(x - 1) + y * xDimension];
			b = board[x + y * xDimension];
			c = board[(x - 1) + (y - 1) * xDimension];
			d = board[x + (y - 1) * xDimension];

			/* check the three special cases first */
			if (a == b && c == d && a == c && b == d) {
				if (plotDots) {
					printf(".");
				} else if (hatchUnassigned && b == -1) {
					printf("\\");
				} else {
					printf(" ");
				}
			} else if (a == b && c == d) {
				printf("-");
			} else if (a == c && b == d) {
				printf("|");
			} else {
				printf("+");
			}

			if (b != d) {
				printf("--");
			} else {
				if (plotDots) {
					printf("..");
				} else if (hatchUnassigned && b == -1) {
					printf("\\\\");
				} else {
					printf("  ");
				}
			}
		}

		/* right, depends on current y and next y */
		x = xDimension - 1;
		b = board[x + y * xDimension];
		d = board[x + (y - 1) * xDimension];

		if (b != d) {
			printf("+");
		} else {
			printf("|");
		}

		printf("\n");
	}

	/* bottom line*/
	y = 0;
	/* contents */
	for (i = 0; i < 2; i++) { /* [size 2] */
		x = 0;
		b = board[x + y * xDimension];

		printf("|"); /* left */

		if (hatchUnassigned && b == -1) {
			printf("\\\\");
		} else {
			printf("  ");
		}

		for (x++; x < xDimension; x++) {
			/* depends on previous x and current x */
			a = board[(x - 1) + y * xDimension];
			b = board[x + y * xDimension];

			if (a != b) {
				printf("|");
			} else if (plotDots) {
				printf(".");
			} else if (hatchUnassigned && b == -1) {
				printf("\\");
			} else {
				printf(" ");
			}

			if (hatchUnassigned && b == -1) {
				printf("\\\\");
			} else {
				printf("  ");
			}
		}

		printf("|"); /* right */
		printf("\n");
	}

	/* bottom left */
	printf("+");
	printf("--");

	for (x = 1; x < xDimension; x++) {
		/* depends on four fields */
		a = board[(x - 1) + y * xDimension];
		b = board[x + y * xDimension];

		if (a != b) {
			printf("+");
		} else {
			printf("-");
		}

		printf("--");
	}

	/* bottom right */
	printf("+");
	printf("\n");
}


/** Stop callback.
*/
static int stop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;

	return 1; /* stop */
}


/** Status callback for 'forward'.
*/
static int printForwardStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;

	/* printf("+"); */

	printf("#");

	return 0; /* don't stop */
}


/** Status callback for 'back'.
*/
static int printBackStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;

	/* printf("-"); */

	/* there ought to be at least one '#' from printForwardStatus();
	* erase it
	*/
	printf("\b \b");

	return 0; /* don't stop */
}


/** Plotting callback.
*/
static int plotter(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension)
{
	(void)tiling;
	(void)figures;
	(void)current;

	plotTiling(board, xDimension, yDimension, 0, 1);

	return 0; /* don't stop */
}

