/* sqtiledr.c -- Square-Figure tiling driver program.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqtiledr.c,v 1.2 2003/03/18 07:59:53 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "sqtile.h"


/* local stuff */
static void demo(int continuous);
static int getNumericalArgument(char const* s);
static void plot(int figureSize, int continuous);
static void plotX(int figureSize, int x, int continuous);
static void usage(void);


/** sqtiledr application starting point.
*/
int main(int argc, char** argv)
{
	int size = -1, x = -1;
	int continuous = 0;
	int i;

	/* option parsing */
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-cont") == 0) {
			continuous = 1;
		} else if (strcmp(argv[i], "-size") == 0) {
			if (i + 1 >= argc) {
				usage();
			}
			size = getNumericalArgument(argv[i + 1]);
			i++;
		} else if (strcmp(argv[i], "-x") == 0) {
			if (i + 1 >= argc) {
				usage();
			}
			x = getNumericalArgument(argv[i + 1]);
			i++;
		} else if (strcmp(argv[i], "-?") == 0) {
			usage();
		} else if (argv[i][0] == '-') {
			usage();
		} else {
			usage();
		}
	}

	if (x >= 0 && size < 0) {
		usage();
	}

	if (size >= 0 && x >= 0) {
		plotX(size, x, continuous);
	} else if (size >= 0) {
		plot(size, continuous);
	} else {
		demo(continuous);
	}

	return 0;
}


/** Usage message.
*/
static void usage(void)
{
	printf(
		"usage: sqtile"
		" [-cont]"
		" [-size [-x]]"
		"\n"
	);
	exit(1);
}


/** Get a numerical argument.  Calls usage() if the argument is
* non-numerical or negative.
*/
static int getNumericalArgument(char const* s)
{
	int n;
	char* p;

	n = strtol(s, &p, 10);
	if (p == s) {
		usage();
	}
	if (n < 0) {
		usage();
	}

	return n;
}


/** Demo.
*/
static void demo(int continuous)
{
	int figureSize;

	for (figureSize = 0; figureSize < MAXFIGURESIZE; figureSize++) {
		/* printf("\n"); */
		plot(figureSize, continuous);
	}
}

/** Plot tilings for a given figure size.
*/
static void plot(int figureSize, int continuous)
{
	figureSet_t const* fs;
	int figures, boardSize, x;

	/* get/generate figures */
	fs = getFigureSet(figureSize);
	figures = DISTINCT_FIGURE_COUNT(fs);
	boardSize = figures * figureSize;

	/* print _all_ productions, ignoring symmetries
	* (e.g. exchanging x/y)
	*/
	for (x = 1; x <= boardSize; x++) {

		/* check if Y will be valid */
		if (boardSize % x != 0) {
			continue;
		}

		/* printf("\n"); */
		plotX(figureSize, x, continuous);
	}
}


/** Plot tilings for a given figure size and X.
*/
static void plotX(int figureSize, int x, int continuous)
{
	figureSet_t const* fs;
	int figures, boardSize, y, *tiling, *board, ok;

	/* get/generate figures */
	fs = getFigureSet(figureSize);
	figures = DISTINCT_FIGURE_COUNT(fs);
	boardSize = figures * figureSize;

	if (x == 0) {
		return;
	}
	if (boardSize % x != 0) {
		return;
	}
	y = boardSize / x;

	/* get the tiling and board buffers */

	tiling = malloc(figures * sizeof(*tiling));
	if (tiling == NULL && figures != 0) {
		return;
	}

	board = malloc(boardSize * sizeof(*board));
	if (board == NULL && boardSize != 0) {
		free(tiling);
		return;
	}

	/* print _all_ productions, ignoring symmetries
	* (four-fold inside the rectangle)
	*/

	/* get and plot tilings, do separators */
	ok = getFirstTiling(fs, x, tiling, board, continuous);
	while (ok) {

		printf("\n");
		plotTiling(board, x, y, 0, 1);

		ok = getNextTiling(fs, x, tiling, board, continuous);
	}

	free(board);
	free(tiling);
}

